/* Copyright (C) 2015-2021 |Meso|Star> (contact@meso-star.com)
 *
 * This software is a library whose purpose is to generate [pseudo] random
 * numbers and random variates.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#ifndef SSP_RNG_C_H
#define SSP_RNG_C_H

#include "ssp.h"
#include <rsys/ref_count.h>

#define SQRT_2_PI 2.50662827463100050240

#ifdef USE_BOOST_RANDOM
  /* Disable some warnings in boost includes */
  #pragma warning(push)
  #pragma warning(disable:4244) /* possible loss of data due to data conversion */
  #pragma warning(disable:4458) /* declaration of a variable hides class member */

  /* The random C++11 library is bugged on MSVC 12 & 14. Use the boost version */
  #include <boost/random.hpp>
  #include <boost/random/random_device.hpp>
  #define RAN_NAMESPACE boost::random

  #pragma warning(pop)
  #pragma warning(push)
  #pragma warning(disable:4706) /* Assignment in conditional expression */
#else
  #include <random>
  #define RAN_NAMESPACE std
#endif

struct ssp_rng {
  struct ssp_rng_type type;
  void* state;
  struct mem_allocator* allocator;

  /* Precomputed RNG constants used to speed up the canonical generations */
  long double r; /* max - min + 1 */
  size_t dbl_k; /* max(1, (#bits_mantisse_double + log2(r)) / log2(r)) */
  size_t flt_k; /* max(1, (#bits_mantisse_float  + log2(r)) / log2(r)) */

  ref_T ref;
};

class rng_cxx /* Wrap a SSP random generator into a CXX object */
{
public:
  FINLINE rng_cxx(struct ssp_rng& _rng) : rng(_rng) {}
  FINLINE uint64_t operator()() { return rng.type.get(rng.state); }
  FINLINE uint64_t min() const { return rng.type.min; }
  FINLINE uint64_t max() const { return rng.type.max; }

  typedef uint64_t result_type;

private:
  struct ssp_rng& rng;
};

#endif /* SSP_RNG_C_H */

